//Scroll16.c (PIC16F628A, 4 MHz, MikroC Pro v5.60)
//Copyright (c) by zuisti, all rights reserved!

//Version 0.1 (1046 words) 46.8 Hz refresh rate
//pure C (best portable but long and very slow !!)
//array[index] access method is used (no pointer)

//version 0.2 (1031 words) 50.4 Hz refresh rate
//the same (pure C) but array access by pointers (shorter and faster)

#define SCROLLTIME 6 // loop count value, the smaller is faster (fastest: 1)

// hw defines:
#define COMMON_CATHODE       //common cathode rows, common anode columns !!!

#define DATA_PIN  RB3_bit //PORTB.3, the hc595 data

//Store pin: PORTA.3  (RA3), positive clock for shiftregs output storage
#define StorePulse RA3_bit = 1; RA3_bit = 0

//clock pin: PORTB.0  (RB0), positive clock for shift in
#define ClockPulse RB0_bit = 1; RB0_bit = 0

#define SCROLL_PIN RB4_bit   // a jumper for (re)start (open) / stop (closed) scrolling

//Line decoder inputs: PORTA<2:0> ;RA0 = LSB
#define DisableLines PORTA = 0b0111  //= 7, all 1 (the outputs are active low)

#include "bitmap7.h" // ASCII font bitmap table, 7 bytes/char

const char message[] = "This is a very-very long message :-) \x0"; // 0 = end sign

//global arrays
char pattbuf[7];   // holds the actual char's line patterns
char linebuf[7*7]; // 7 bytes for the first (uppermost) line (row), 7 bytes for the 2nd, ...

void HW_Init(void) {
    CMCON = 0b00000111; // switch off comparators
    DisableLines;       // PORTA = 7, also store pin is zeroed
    PORTB = 0;          // also data and clock pins are zeroed
    TRISA = 0;          // PORTA pins are outputs
    TRISB = 0xF2;       // only PORTB.0,2,3 are outputs
    NOT_RBPU_bit = 0;   // portb pullups on

/* init timer2 for line timing
    one row display = 861 us (without timer2 waiting) but
    with scroll (worst case) (10+38+369+1516 = 1.933 ms)
    sum = 1933 + 861 = 2794 us (min. for timer2)

    Prescaler=1:1; PostScaler=1:12; Period = 2.808 ms
    4 MHz: 2.808 ms/line, 19.84 ms the whole loop (50.4 Hz refresh rate)
*/
    PR2 = 234;           // Timer2 Match value
    T2CON = 0b01011100;  // enable timer2 with 1:12 postscaler, 1:1 prescaler
    TMR2IF_bit = 0;      // clear IT flag
    PIE1 = 2; // only TMR2IE_bit = 1; enable Timer2 IT

    INTCON = 0xC0; // Enable global (GIE_bit) and periferial (PEIE_bit) ITs
}

void interrupt(void) {
    DisableLines;   // PORTA = 7
    TMR2ON_bit = 0; // disable timer2
    TMR2IF_bit = 0; // clear IT flag
}

void get_patterns(char ch) {
     const char *bitmap_ptr;
     char *patptr, rowcnt;

     patptr = pattbuf;
     ch -= 32; // the space is the first in table
     //calculate the table address of the first pattern
     bitmap_ptr = bitmap + (8*(char)ch) - ch; // = bitmap + 7*ch, but faster

     rowcnt = 7; //we have 7 patterns
     do {
        ch = *bitmap_ptr++; // get it from bitmap table
        ch <<= 2; // drop unused two bits (7,6)
#ifdef COMMON_CATHODE
        ch = ~ch; // inverted (0<>1)
#endif
        *patptr++ = ch; // put to the ram array
     } while (--rowcnt);
}

char nextchar(void) {
     char ch;
     static const char *msg_ptr = message;

     for (;;) {
         ch = *msg_ptr++;
         if (ch)
            break;              // not 0 (end sign)
         msg_ptr = message;     // else back to begin
     }
     return ch;
}
     
void shift_in_a_column(void) {
     char *pat_ptr, *buf_ptr, ch, rowcnt, bytecnt, carry_over;

     pat_ptr = pattbuf + 6;  // last (7th) pattern
     buf_ptr = linebuf + 48; // last line's last byte

     rowcnt = 7;
     do {
        // unfortunately (in MikroC) no rotate with carry

        // pattern rotate left with carry
        ch = *pat_ptr;
        carry_over = ch & 0x80;
        ch <<= 1; // a 0 shifted in from right but not used
        *pat_ptr-- = ch;

        // rotate a whole line (7 bytes) left with carry
        bytecnt = 7;
        do {
           ch = *buf_ptr;
           ch <<= 1; // carry = old bit7, new bit0 = 0
           if (carry_over) ch.B0 = 1;// carry NOT destroyed
           carry_over = STATUS & 1;  // value of the carry flag ! this is the repair
           *buf_ptr-- = ch;          // carry IS destroyed here!
           //carry_over = ch & 0x80; // fixed and moved up
        } while (--bytecnt);
     } while (--rowcnt);
}

void display(void) {
   char *buf_ptr, ch, bytecnt, bitcnt, rowcnt;

   buf_ptr = linebuf; //first byte of buffer
   rowcnt = 0; // loop until rowcnt = 7
   do { // 861 us / row
      bytecnt = 7;
      do {
         ch = *buf_ptr++;
         bitcnt = 8;
         do {
            DATA_PIN = ch.B7;
            ch <<= 1;
            ClockPulse;
         } while (--bitcnt);
      }while (--bytecnt);

      //timed display (same row brightness)
      do {} while (TMR2ON_bit); //wait for timer2 IT (timer2 and all lines off)
      TMR2ON_bit = 1; //restart timer2, also a time gap to switch off drivers

      StorePulse;
      PORTA = rowcnt; // switch on the actual line driver
   } while (++rowcnt != 7);
}

void clear_linebuf(void) {
  char *bufptr, cnt;

  bufptr = linebuf + 48; // last byte
  cnt = 49;
  do {
#ifdef COMMON_CATHODE
     *bufptr-- = 0;
#else
     *bufptr-- = 255;
#endif
  } while (--cnt);
}

void main(void) {
  char width, scrollcnt, ch;
  
  HW_Init();
  clear_linebuf();
  //for prompt scroll:
  width = 0;
  scrollcnt = 1; 

  for (;;) {
    if (SCROLL_PIN && --scrollcnt == 0) {
       scrollcnt = SCROLLTIME;
       if (!width) {
          ch = nextchar();  //38 us
          width = 6;
          if (ch == 32) width = 3;
          get_patterns(ch); //369 us
       }
       --width;
       shift_in_a_column(); //1516 us !!!!
    }
    display(); //6.03 ms (without timer2 waiting !) 6.03/7 = appr 861 us / row
  }
}
//------ end