//Scroll16.c (PIC16F628A, 4 MHz, MikroC Pro v5.60)
//Copyright (c) by zuisti, all rights reserved!

//Version 0.1 (1046 words) 46.8 Hz refresh rate
//pure C (best portable and best to understand but long and very slow !!)
//array[index] access method is used (no pointer)

//version 0.2 (1031 words) 50.4 Hz refresh rate
//the same (pure C) but array access by pointers (shorter and faster a bit)

//version 0.3 (1010 words) 60.4 Hz refresh rate (a significant improvement)
//the same as 0.2 (pure C) but indirect array access by FSR and INDF (legal in MikroC)
//FSR is a char pointer (hardware char *), and INDF is the *FSR (hw too)
// only one FSR exists in a PIC16 (except the enhanced types) but three in PIC18


#define SCROLLTIME 6 // loop count value, the smaller is faster (fastest: 1)

// hw defines:
#define COMMON_CATHODE       //common cathode rows, common anode columns !!!

#define DATA_PIN  RB3_bit //PORTB.3, the hc595 data

//Store pin: PORTA.3  (RA3), positive clock for shiftregs output storage
#define StorePulse RA3_bit = 1; RA3_bit = 0

//clock pin: PORTB.0  (RB0), positive clock for shift in
#define ClockPulse RB0_bit = 1; RB0_bit = 0

#define SCROLL_PIN RB4_bit   // a jumper for (re)start (open) / stop (closed) scrolling

//Line decoder inputs: PORTA<2:0> ;RA0 = LSB
#define DisableLines PORTA = 0b0111  //= 7, all 1 (the outputs are active low)

#include "bitmap7.h" // ASCII font bitmap table, 7 bytes/char

const char message[] = "This is a very-very long message :-) \x0"; // 0 = end sign

//global arrays
char pattbuf[7];   // holds the actual char's line patterns
char linebuf[7*7]; // 7 bytes for the first (uppermost) line (row), 7 bytes for the 2nd, ...

void HW_Init(void) {
    CMCON = 0b00000111; // switch off comparators
    DisableLines;       // PORTA = 7, also store pin is zeroed
    PORTB = 0;          // also data and clock pins are zeroed
    TRISA = 0;          // PORTA pins are outputs
    TRISB = 0xF2;       // only PORTB.0,2,3 are outputs
    NOT_RBPU_bit = 0;   // portb pullups on

/* init timer2 for line timing
    one row display = 849 us (without timer2 waiting) but
    with scroll (worst case) (10+38+334+1096 = 1.478 ms)
    sum = 1478 + 849 = 2327 us (min. for timer2)

    Prescaler=1:1; PostScaler=1:12; Period = 2.340 ms
    4 MHz: 2.340 ms/line, 16.55 ms the whole loop (60.4 Hz refresh rate)
*/
    PR2 = 195;           // Timer2 Match value
    T2CON = 0b01011100;  // enable timer2 with 1:12 postscaler, 1:1 prescaler
    TMR2IF_bit = 0;      // clear IT flag
    PIE1 = 2; // only TMR2IE_bit = 1; enable Timer2 IT

    INTCON = 0xC0; // Enable global (GIE_bit) and periferial (PEIE_bit) ITs
}

void interrupt(void) {
    DisableLines;   // PORTA = 7
    TMR2ON_bit = 0; // disable timer2
    TMR2IF_bit = 0; // clear IT flag
}

void get_patterns(char ch) {
     const char *bitmap_ptr;
     char rowcnt;

     FSR = (char)pattbuf;
     ch -= 32; // the space is the first in table
     //calculate the table address of the first pattern
     bitmap_ptr = bitmap + (8*(char)ch) - ch; // = bitmap + 7*ch, but faster

     rowcnt = 7; //we have 7 patterns
     do {
        ch = *bitmap_ptr++; // get it from bitmap table
        ch <<= 2; // drop unused two bits (7,6)
        // put it to array
#ifdef COMMON_CATHODE
        INDF = ~ch; // inverted (0<>1)
#else
        INDF = ch;
#endif
        ++FSR;
     } while (--rowcnt);
}

char nextchar(void) {
     char ch;
     static const char *msg_ptr = message;

     for (;;) {
         ch = *msg_ptr++;
         if (ch)
            break;              // not 0 (end sign)
         msg_ptr = message;     // else back to begin
     }
     return ch;
}
     
void shift_in_a_column(void) {
     char pat_fsr, buf_fsr, rowcnt, bytecnt, carry_over;

     pat_fsr = (char)(pattbuf + 6);  // last (7th) pattern
     buf_fsr = (char)(linebuf + 48); // global, points the last line's last byte
     //because there is only one FSR

     rowcnt = 7;
     do {
        // first pattern rotate left, carry to carry_over
        FSR = pat_fsr--;
        carry_over = INDF & 0x80; // bit7
        INDF <<= 1;      // a 0 shifted in from right but not used
        
        // then rotate a whole line (7 bytes from buf_fsr), FSR is destroyed
        FSR = buf_fsr;     // first points the buffer end
        bytecnt = 7;
        do {
           INDF <<= 1; // carry = old bit7, new bit0 = 0
           if (carry_over) INDF.B0 = 1; // carry NOT destroyed
           carry_over = STATUS & 1;     // value of the carry flag ! this is the repair
           --FSR;
        } while (--bytecnt);
        buf_fsr = FSR;
     } while (--rowcnt);
}

void display(void) {
   char ch, bytecnt, bitcnt, rowcnt;

   FSR = (char)linebuf; // first byte
   rowcnt = 0; // loop until rowcnt = 7
   do {
      bytecnt = 7;  //839 us
      do {
         ch = INDF;
         bitcnt = 8;
         do {
            DATA_PIN = ch.B7;
            ch <<= 1;
            ClockPulse;
         } while (--bitcnt);
         FSR++;
      }while (--bytecnt);

      //timed display (same row brightness)
      do {} while (TMR2ON_bit); //wait for timer2 IT (timer2 and all lines off)
      TMR2ON_bit = 1; //restart timer2, also a time gap to switch off drivers

      StorePulse;
      PORTA = rowcnt; // switch on the actual line driver
   } while (++rowcnt != 7);
}

void clear_linebuf(void) {
  char cnt = 49;

  FSR = (char)(linebuf + 48);
  do {
#ifdef COMMON_CATHODE
     INDF = 0;
#else
     INDF = 255;
#endif
     --FSR;
  } while (--cnt);
}

void main(void) {
  char width, scrollcnt, ch;
  
  HW_Init();
  clear_linebuf();
  //for prompt scroll:
  width = 0;
  scrollcnt = 1; 

  for (;;) {
    if (SCROLL_PIN && --scrollcnt == 0) {
       scrollcnt = SCROLLTIME;
       if (!width) {
          ch = nextchar();  //38, 38 us
          width = 6;
          if (ch == 32) width = 3;
          get_patterns(ch); //334, 369 us
       }
       --width;
       shift_in_a_column(); //1096, 1516 us !!!!
    }
    display(); //5.94 ms (without timer2 waiting !) 5.94/7 = appr 849 us / row
  }
}
//------ end